@@ -217,28 +217,45 @@ impl Default for Caching {
     }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 struct CacheEntry {
     filter: LevelFilter,
     logger: PyObject,
 }
 
-#[derive(Clone, Debug, Default)]
+impl CacheEntry {
+    fn clone_ref(&self, py: Python<'_>) -> Self {
+        CacheEntry {
+            filter: self.filter,
+            logger: self.logger.clone_ref(py),
+        }
+    }
+}
+
+#[derive(Debug, Default)]
 struct CacheNode {
     local: Option<CacheEntry>,
     children: HashMap<String, Arc<CacheNode>>,
 }
 
 impl CacheNode {
-    fn store_to_cache_recursive<'a, P>(&self, mut path: P, entry: CacheEntry) -> Arc<Self>
+    fn store_to_cache_recursive<'a, P>(
+        &self,
+        py: Python<'_>,
+        mut path: P,
+        entry: CacheEntry,
+    ) -> Arc<Self>
     where
         P: Iterator<Item = &'a str>,
     {
-        let mut me = self.clone();
+        let mut me = CacheNode {
+            children: self.children.clone(),
+            local: self.local.as_ref().map(|e| e.clone_ref(py)),
+        };
         match path.next() {
             Some(segment) => {
                 let child = me.children.entry(segment.to_owned()).or_default();
-                *child = child.store_to_cache_recursive(path, entry);
+                *child = child.store_to_cache_recursive(py, path, entry);
             }
             None => me.local = Some(entry),
         }
@@ -253,7 +270,7 @@ impl CacheNode {
 ///
 /// It can be either created directly and then installed, passed to other aggregating log systems,
 /// or the [`init`] or [`try_init`] functions may be used if defaults are good enough.
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 pub struct Logger {
     /// Filter used as a fallback if none of the `filters` match.
     top_filter: LevelFilter,
@@ -286,7 +303,7 @@ impl Logger {
     ///
     /// It defaults to having a filter for [`Debug`][LevelFilter::Debug].
     pub fn new(py: Python<'_>, caching: Caching) -> PyResult<Self> {
-        let logging = py.import_bound("logging")?;
+        let logging = py.import("logging")?;
         Ok(Self {
             top_filter: LevelFilter::Debug,
             filters: HashMap::new(),
@@ -422,8 +439,8 @@ impl Logger {
                     record.file(),
                     record.line().unwrap_or_default(),
                     msg,
-                    PyTuple::empty_bound(py), // args
-                    &none,                    // exc_info
+                    PyTuple::empty(py), // args
+                    &none,              // exc_info
                 ),
             )?;
             logger.call_method1("handle", (record,))?;
@@ -464,12 +481,12 @@ impl Logger {
         metadata.level() <= cache_filter && metadata.level() <= self.filter_for(metadata.target())
     }
 
-    fn store_to_cache(&self, target: &str, entry: CacheEntry) {
+    fn store_to_cache(&self, py: Python<'_>, target: &str, entry: CacheEntry) {
         let path = target.split("::");
 
         let orig = self.cache.load();
         // Construct a new cache structure and insert the new root.
-        let new = orig.store_to_cache_recursive(path, entry);
+        let new = orig.store_to_cache_recursive(py, path, entry);
         // Note: In case of collision, the cache update is lost. This is fine, as we simply lose a
         // tiny bit of performance and will cache the thing next time.
         //
@@ -497,30 +514,41 @@ impl Log for Logger {
     fn log(&self, record: &Record) {
         let cache = self.lookup(record.target());
 
-        let mut store_to_cache = None;
         if self.enabled_inner(record.metadata(), &cache) {
-            Python::with_gil(|py| match self.log_inner(py, record, &cache) {
-                Ok(Some(logger)) => {
-                    let filter = match self.caching {
-                        Caching::Nothing => unreachable!(),
-                        Caching::Loggers => LevelFilter::max(),
-                        Caching::LoggersAndLevels => extract_max_level(logger.bind(py))
-                            .unwrap_or_else(|e| {
-                                e.print(py);
-                                LevelFilter::max()
-                            }),
-                    };
-                    store_to_cache = Some((logger, filter));
-                }
-                Ok(None) => (),
-                Err(e) => e.print(py),
-            })
-        }
-        // Note: no more GIL here. Not needed for storing to cache.
+            Python::with_gil(|py| {
+                // If an exception were triggered before this attempt to log,
+                // store it to the side for now and restore it afterwards.
+                let maybe_existing_exception = PyErr::take(py);
+                match self.log_inner(py, record, &cache) {
+                    Ok(Some(logger)) => {
+                        let filter = match self.caching {
+                            Caching::Nothing => unreachable!(),
+                            Caching::Loggers => LevelFilter::max(),
+                            Caching::LoggersAndLevels => extract_max_level(logger.bind(py))
+                                .unwrap_or_else(|e| {
+                                    // See detailed NOTE below
+                                    e.restore(py);
+                                    LevelFilter::max()
+                                }),
+                        };
 
-        if let Some((logger, filter)) = store_to_cache {
-            let entry = CacheEntry { filter, logger };
-            self.store_to_cache(record.target(), entry);
+                        let entry = CacheEntry { filter, logger };
+                        self.store_to_cache(py, record.target(), entry);
+                    }
+                    Ok(None) => (),
+                    Err(e) => {
+                        // NOTE: If an exception was triggered _during_ logging, restore it as current Python exception.
+                        // We have to use PyErr::restore because we cannot return a PyResult from the Log trait's log method.
+                        e.restore(py);
+                    }
+                };
+
+                // If there was a prior exception, restore it now
+                // This ensures that the earliest thrown exception will be the one that's visible to the caller.
+                if let Some(e) = maybe_existing_exception {
+                    e.restore(py);
+                }
+            })
         }
     }
 
